@@ -1,7 +1,7 @@
 /*
  * Dynamic Lua binding to GObject using dynamic gobject-introspection.
  *
- * Copyright (c) 2010, 2011, 2012 Pavel Holejsovsky
+ * Copyright (c) 2010, 2011, 2012, 2013 Pavel Holejsovsky
  * Licensed under the MIT license:
  * http://www.opensource.org/licenses/mit-license.php
  *
@@ -11,6 +11,32 @@
 #include <string.h>
 #include "lgi.h"
 
+/* GLib 2.32 deprecated GStaticRecMutex in favor of GRecMutex.  For
+   older GLib versions, use still older version. */
+#if !GLIB_CHECK_VERSION(2, 32, 0)
+#define GRecMutex GStaticRecMutex
+#define G_REC_MUTEX_INIT  = G_STATIC_REC_MUTEX_INIT
+#define g_rec_mutex_init g_static_rec_mutex_init
+#define g_rec_mutex_lock g_static_rec_mutex_lock
+#define g_rec_mutex_unlock g_static_rec_mutex_unlock
+#define g_rec_mutex_clear g_static_rec_mutex_free
+#else
+#define G_REC_MUTEX_INIT
+#endif
+
+/* GLib 2.30 changed g_atomic_int_add() to return the new value. For
+   older GLib versions, use g_atomic_int_exchange_and_add() which did. */
+#if !GLIB_CHECK_VERSION(2, 30, 0)
+#ifdef g_atomic_int_add
+#undef g_atomic_int_add
+#endif
+
+#define g_atomic_int_add(atomic, val) \
+  (g_atomic_int_exchange_and_add ((atomic), (val)))
+#endif
+
+volatile gint global_state_id = 0;
+
 #ifndef NDEBUG
 const char *lgi_sd (lua_State *L)
 {
@@ -97,6 +123,10 @@ lgi_type_get_name (lua_State *L, GIBaseInfo *info)
   int n = 1;
   lua_pushstring (L, g_base_info_get_namespace (info));
 
+  if (g_base_info_get_type (info) == GI_INFO_TYPE_CALLBACK)
+    /* Avoid duplicate name for callbacks. */
+    info = g_base_info_get_container (info);
+
   /* Add names on the whole path, but in reverse order. */
   for (; info != NULL; info = g_base_info_get_container (info))
     if (!GI_IS_TYPE_INFO (info))
@@ -147,7 +177,7 @@ lgi_type_get_repotype (lua_State *L, GType gtype, GIBaseInfo *info)
     {
       /* Not indexed yet.  Try to lookup by name - this works when
 	 lazy-loaded repo tables are not loaded yet. */
-      if (!info)
+      if (!info && gtype != G_TYPE_INVALID)
 	{
 	  info = g_irepository_find_by_gtype (NULL, gtype);
 	  lgi_gi_info_new (L, info);
@@ -266,7 +296,8 @@ core_constant (lua_State *L)
   GITypeInfo *ti = g_constant_info_get_type (ci);
   lgi_gi_info_new (L, ti);
   g_constant_info_get_value (ci, &val);
-  lgi_marshal_2lua (L, ti, GI_TRANSFER_NOTHING, &val, 0, NULL, NULL);
+  lgi_marshal_2lua (L, ti, NULL, GI_DIRECTION_IN, GI_TRANSFER_NOTHING, &val,
+		    0, NULL, NULL);
   return 1;
 }
 
@@ -274,21 +305,21 @@ typedef struct _LgiStateMutex
 {
   /* Pointer to either local state lock (next member of this
      structure) or to global package lock. */
-  GStaticRecMutex *mutex;
-  GStaticRecMutex state_mutex;
+  GRecMutex *mutex;
+  GRecMutex state_mutex;
 } LgiStateMutex;
 
 /* Global package lock (the one used for
    gdk_threads_enter/clutter_threads_enter) */
-static GStaticRecMutex package_mutex = G_STATIC_REC_MUTEX_INIT;
+static GRecMutex package_mutex G_REC_MUTEX_INIT;
 
-/* GC method for GStaticRecMutex structure, which lives inside lua_State. */
+/* GC method for GRecMutex structure, which lives inside lua_State. */
 static int
 call_mutex_gc (lua_State* L)
 {
   LgiStateMutex *mutex = lua_touserdata (L, 1);
-  g_static_rec_mutex_unlock (mutex->mutex);
-  g_static_rec_mutex_free (&mutex->state_mutex);
+  g_rec_mutex_unlock (mutex->mutex);
+  g_rec_mutex_clear (&mutex->state_mutex);
   return 0;
 }
 
@@ -314,7 +345,7 @@ void
 lgi_state_enter (gpointer state_lock)
 {
   LgiStateMutex *mutex = state_lock;
-  GStaticRecMutex *wait_on;
+  GRecMutex *wait_on;
 
   /* There is a complication with lock switching.  During the wait for
      the lock, someone could call core.registerlock() and thus change
@@ -322,12 +353,12 @@ lgi_state_enter (gpointer state_lock)
   for (;;)
     {
       wait_on = g_atomic_pointer_get (&mutex->mutex);
-      g_static_rec_mutex_lock (wait_on);
+      g_rec_mutex_lock (wait_on);
       if (wait_on == mutex->mutex)
 	break;
 
       /* The lock is changed, unlock this one and wait again. */
-      g_static_rec_mutex_unlock (wait_on);
+      g_rec_mutex_unlock (wait_on);
     }
 }
 
@@ -336,7 +367,7 @@ lgi_state_leave (gpointer state_lock)
 {
   /* Get pointer to the call mutex belonging to this state. */
   LgiStateMutex *mutex = state_lock;
-  g_static_rec_mutex_unlock (mutex->mutex);
+  g_rec_mutex_unlock (mutex->mutex);
 }
 
 static const char* log_levels[] = {
@@ -349,7 +380,14 @@ core_log (lua_State *L)
   const char *domain = luaL_checkstring (L, 1);
   int level = 1 << (luaL_checkoption (L, 2, log_levels[5], log_levels) + 2);
   const char *message = luaL_checkstring (L, 3);
+
+#if GLIB_CHECK_VERSION(2, 50, 0)
+  /* TODO: We can include more debug information such as lua line numbers */
+  g_log_structured (domain, level, "MESSAGE", "%s", message);
+#else
   g_log (domain, level, "%s", message);
+#endif
+
   return 0;
 }
 
@@ -369,13 +407,13 @@ core_yield (lua_State *L)
 static void
 package_lock_enter (void)
 {
-  g_static_rec_mutex_lock (&package_mutex);
+  g_rec_mutex_lock (&package_mutex);
 }
 
 static void
 package_lock_leave (void)
 {
-  g_static_rec_mutex_unlock (&package_mutex);
+  g_rec_mutex_unlock (&package_mutex);
 }
 
 static gpointer package_lock_register[8] = { NULL };
@@ -385,7 +423,7 @@ core_registerlock (lua_State *L)
 {
   void (*set_lock_functions)(GCallback, GCallback);
   LgiStateMutex *mutex;
-  GStaticRecMutex *wait_on;
+  GRecMutex *wait_on;
   unsigned i;
 
   /* Get registration function. */
@@ -413,9 +451,9 @@ core_registerlock (lua_State *L)
   wait_on = g_atomic_pointer_get (&mutex->mutex);
   if (wait_on != &package_mutex)
     {
-      g_static_rec_mutex_lock (&package_mutex);
+      g_rec_mutex_lock (&package_mutex);
       g_atomic_pointer_set (&mutex->mutex, &package_mutex);
-      g_static_rec_mutex_unlock (wait_on);
+      g_rec_mutex_unlock (wait_on);
     }
   return 0;
 }
@@ -443,6 +481,10 @@ module_gc (lua_State *L)
 {
   GModule **module = luaL_checkudata (L, 1, UD_MODULE);
   g_module_close (*module);
+
+  /* Unset the metatable / make the module unusable */
+  lua_pushnil (L);
+  lua_setmetatable (L, 1);
   return 0;
 }
 
@@ -468,12 +510,15 @@ static const struct luaL_Reg module_reg[] = {
   { NULL, NULL }
 };
 
-#ifdef G_WITH_CYGWIN
+#if defined(G_WITH_CYGWIN)
 #define MODULE_NAME_FORMAT_VERSION "cyg%s-%d.dll"
 #define MODULE_NAME_FORMAT_PLAIN "cyg%s.dll"
-#elif G_OS_WIN32
+#elif defined(G_OS_WIN32)
 #define MODULE_NAME_FORMAT_VERSION "lib%s-%d.dll"
 #define MODULE_NAME_FORMAT_PLAIN "lib%s.dll"
+#elif defined(__APPLE__)
+#define MODULE_NAME_FORMAT_VERSION "lib%s.%d.dylib"
+#define MODULE_NAME_FORMAT_PLAIN "lib%s.dylib"
 #else
 #define MODULE_NAME_FORMAT_VERSION "lib%s.so.%d"
 #define MODULE_NAME_FORMAT_PLAIN "lib%s.so"
@@ -487,15 +532,28 @@ core_module (lua_State *L)
 {
   char *name;
 
-  /* If the version is present, combine it with basename. */
+  /* If the version is present, combine it with basename.
+     Except on OpenBSD, where libraries are versioned like libfoo.so.0.0
+     and we will always load the shared object with the highest version
+     number.
+   */
+#ifndef __OpenBSD__
   if (!lua_isnoneornil (L, 2))
     name = g_strdup_printf (MODULE_NAME_FORMAT_VERSION,
 			    luaL_checkstring (L, 1),
 			    (int) luaL_checkinteger (L, 2));
   else
+#endif
     name = g_strdup_printf (MODULE_NAME_FORMAT_PLAIN,
 			    luaL_checkstring (L, 1));
 
+#if defined(__APPLE__)
+  char *path = g_module_build_path (GOBJECT_INTROSPECTION_LIBDIR,
+                                    name);
+  g_free(name);
+  name = path;
+#endif
+
   /* Try to load the module. */
   GModule *module = g_module_open (name, 0);
   if (module == NULL)
@@ -515,6 +573,22 @@ core_module (lua_State *L)
   return 2;
 }
 
+static int core_upcase (lua_State *L)
+{
+  gchar *str = g_ascii_strup (luaL_checkstring (L, 1), -1);
+  lua_pushstring (L, str);
+  g_free (str);
+  return 1;
+}
+
+static int core_downcase (lua_State *L)
+{
+  gchar *str = g_ascii_strdown (luaL_checkstring (L, 1), -1);
+  lua_pushstring (L, str);
+  g_free (str);
+  return 1;
+}
+
 static const struct luaL_Reg lgi_reg[] = {
   { "log",  core_log },
   { "gtype", core_gtype },
@@ -525,6 +599,8 @@ static const struct luaL_Reg lgi_reg[] = {
   { "band", core_band },
   { "bor", core_bor },
   { "module", core_module },
+  { "upcase", core_upcase },
+  { "downcase", core_downcase },
   { NULL, NULL }
 };
 
@@ -563,6 +639,17 @@ set_resident (lua_State *L)
     }
   else
     {
+      if (lua_gettop(L) == 3)
+	{
+	  /* Some Lua versions give us the path to the .so on the stack.
+	     Just load & leak it. */
+	  GModule* module = g_module_open(lua_tostring(L, 2),
+					  G_MODULE_BIND_LAZY |
+					  G_MODULE_BIND_LOCAL);
+	  if (module != NULL)
+	    return;
+	}
+
       /* This hack tries to enumerate the whole registry table and
 	 find 'LOADLIB: path' library.  When it detects itself, it
 	 just removes pointer to the loaded library, disallowing Lua
@@ -596,23 +683,27 @@ set_resident (lua_State *L)
     }
 }
 
-int
+G_MODULE_EXPORT int
 luaopen_lgi_corelgilua51 (lua_State* L)
 {
   LgiStateMutex *mutex;
+  gint state_id;
 
   /* Try to make itself resident.  This is needed because this dynamic
      module is 'statically' linked with glib/gobject, and these
      libraries are not designed to be unloaded.  Once they are
      unloaded, they cannot be safely loaded again into the same
-     process.  To avoid problems when repeately opening and closing
+     process.  To avoid problems when repeatedly opening and closing
      lua_States and loading lgi into them, we try to make the whole
      'core' module resident. */
   set_resident (L);
 
+#if !GLIB_CHECK_VERSION(2, 36, 0)
+  g_type_init ();
+#endif
+
   /* Early GLib initializations. Make sure that following fundamental
      G_TYPEs are already initialized. */
-  g_type_init ();
   volatile GType unused;
   unused = G_TYPE_DATE;
   unused = G_TYPE_REGEX;
@@ -645,8 +736,8 @@ luaopen_lgi_corelgilua51 (lua_State* L)
   lua_pushlightuserdata (L, &call_mutex);
   mutex = lua_newuserdata (L, sizeof (*mutex));
   mutex->mutex = &mutex->state_mutex;
-  g_static_rec_mutex_init (&mutex->state_mutex);
-  g_static_rec_mutex_lock (&mutex->state_mutex);
+  g_rec_mutex_init (&mutex->state_mutex);
+  g_rec_mutex_lock (&mutex->state_mutex);
   lua_pushlightuserdata (L, &call_mutex_mt);
   lua_rawget (L, LUA_REGISTRYINDEX);
   lua_setmetatable (L, -2);
@@ -656,6 +747,22 @@ luaopen_lgi_corelgilua51 (lua_State* L)
   lua_newtable (L);
   luaL_register (L, NULL, lgi_reg);
 
+  /* Add the state ID */
+  state_id = g_atomic_int_add (&global_state_id, 1);
+  if (state_id == 0)
+    lua_pushliteral (L, "");
+  else
+    lua_pushfstring (L, "+L%d", state_id);
+  lua_setfield (L, -2, "id");
+
+  /* Add lock and enter/leave locking functions. */
+  lua_pushlightuserdata (L, lgi_state_get_lock (L));
+  lua_setfield (L, -2, "lock");
+  lua_pushlightuserdata (L, lgi_state_enter);
+  lua_setfield (L, -2, "enter");
+  lua_pushlightuserdata (L, lgi_state_leave);
+  lua_setfield (L, -2, "leave");
+
   /* Create repo and index table. */
   create_repo_table (L, "index", &repo_index);
   create_repo_table (L, "repo", &repo);
