@@ -39,6 +39,7 @@ function testsuite.group:run(id)
       self.results.total = self.results.total + 1
       if testsuite.verbose then
 	 io.write(('%-8s:%3d:%-35s'):format(self.name, num, self[num]))
+	 io.flush()
       end
       local ok, msg
       local func = self[self[num]]
@@ -110,9 +111,13 @@ for _, sourcefile in ipairs {
    'corocbk.lua',
    'record.lua',
    'gobject.lua',
+   'glib.lua',
    'variant.lua',
+   'dbus.lua',
    'gtk.lua',
    'cairo.lua',
+   'pango.lua',
+   'gio.lua',
 } do
    dofile(testpath .. '/' .. sourcefile)
 end
